@@ -6,9 +6,9 @@ import os
 import W
 import WASTEconst
 import PyBrowser
-import Qd
-import Evt
-import Lists
+from Carbon import Qd
+from Carbon import Evt
+from Carbon import Lists
 import MacOS
 _filenames = {}
 
@@ -65,7 +65,7 @@ class Debugger(bdb.Bdb):
 	
 	def start(self, bottomframe = None, running = 0):
 		W.getapplication().DebuggerQuit = bdb.BdbQuit
-		import Menu
+		from Carbon import Menu
 		Menu.HiliteMenu(0)
 		if self.closed:
 			self.setupwidgets(self.title)
@@ -681,7 +681,6 @@ class BreakpointsViewer:
 	
 	def __init__(self, debugger):
 		self.debugger = debugger
-		import Lists
 		self.w = W.Window((300, 250), 'Breakpoints', minsize = (200, 200))
 		self.w.panes = W.HorizontalPanes((8, 8, -8, -32), (0.3, 0.7))
 		self.w.panes.files = W.List(None, callback = self.filehit)		#, flags = Lists.lOnlyOne)
