@@ -2,6 +2,7 @@
 /* Time module */
 
 #include "Python.h"
+#include "structseq.h"
 
 #include <ctype.h>
 
@@ -16,9 +17,7 @@
 #define GUSI_TO_MSL_EPOCH (4*365*24*60*60)
 #endif /* USE_GUSI2 */
 #else
-#ifndef RISCOS
 #include <sys/types.h>
-#endif /* RISCOS */
 #endif
 
 #ifdef QUICKWIN
@@ -88,22 +87,22 @@ static PyObject *moddict;
 */
 static long timezone;
 
-static void 
+static void
 initmactimezone(void)
 {
 	MachineLocation	loc;
 	long		delta;
 
 	ReadLocation(&loc);
-	
+
 	if (loc.latitude == 0 && loc.longitude == 0 && loc.u.gmtDelta == 0)
 		return;
-	
+
 	delta = loc.u.gmtDelta & 0x00FFFFFF;
-	
+
 	if (delta & 0x00800000)
 		delta |= 0xFF000000;
-	
+
 	timezone = -delta;
 }
 #endif /* macintosh */
@@ -162,10 +161,10 @@ time_clock(PyObject *self, PyObject *args)
 
 	if (LargeIntegerEqualToZero(divisor)) {
 		QueryPerformanceCounter(&ctrStart);
-		if (!QueryPerformanceFrequency(&divisor) || 
+		if (!QueryPerformanceFrequency(&divisor) ||
 		    LargeIntegerEqualToZero(divisor)) {
-				/* Unlikely to happen - 
-				   this works on all intel machines at least! 
+				/* Unlikely to happen -
+				   this works on all intel machines at least!
 				   Revert to clock() */
 			return PyFloat_FromDouble(clock());
 		}
@@ -174,10 +173,10 @@ time_clock(PyObject *self, PyObject *args)
 	diff = LargeIntegerSubtract(now, ctrStart);
 	diff = LargeIntegerDivide(diff, divisor, &rem);
 	/* XXX - we assume both divide results fit in 32 bits.  This is
-	   true on Intels.  First person who can afford a machine that 
+	   true on Intels.  First person who can afford a machine that
 	   doesnt deserves to fix it :-)
 	*/
-	return PyFloat_FromDouble((double)diff.LowPart + 
+	return PyFloat_FromDouble((double)diff.LowPart +
 		              ((double)rem.LowPart / (double)divisor.LowPart));
 }
 
@@ -210,19 +209,53 @@ static char sleep_doc[] =
 Delay execution for a given number of seconds.  The argument may be\n\
 a floating point number for subsecond precision.";
 
+static PyStructSequence_Field struct_time_type_fields[] = {
+	{"tm_year", NULL},
+	{"tm_mon", NULL},
+	{"tm_mday", NULL},
+	{"tm_hour", NULL},
+	{"tm_min", NULL},
+	{"tm_sec", NULL},
+	{"tm_wday", NULL},
+	{"tm_yday", NULL},
+	{"tm_isdst", NULL},
+	{0}
+};
+
+static PyStructSequence_Desc struct_time_type_desc = {
+	"time.struct_time",
+	NULL,
+	struct_time_type_fields,
+	9,
+};
+	
+static PyTypeObject StructTimeType;
+
 static PyObject *
 tmtotuple(struct tm *p)
 {
-	return Py_BuildValue("(iiiiiiiii)",
-			     p->tm_year + 1900,
-			     p->tm_mon + 1,	   /* Want January == 1 */
-			     p->tm_mday,
-			     p->tm_hour,
-			     p->tm_min,
-			     p->tm_sec,
-			     (p->tm_wday + 6) % 7, /* Want Monday == 0 */
-			     p->tm_yday + 1,	   /* Want January, 1 == 1 */
-			     p->tm_isdst);
+	PyObject *v = PyStructSequence_New(&StructTimeType);
+	if (v == NULL)
+		return NULL;
+	
+#define SET(i,val) PyStructSequence_SET_ITEM(v, i, PyInt_FromLong((long) val))
+
+	SET(0, p->tm_year + 1900);
+	SET(1, p->tm_mon + 1);	   /* Want January == 1 */
+	SET(2, p->tm_mday);
+	SET(3, p->tm_hour);
+	SET(4, p->tm_min);
+	SET(5, p->tm_sec);
+	SET(6, (p->tm_wday + 6) % 7); /* Want Monday == 0 */
+	SET(7, p->tm_yday + 1);	   /* Want January, 1 == 1 */
+	SET(8, p->tm_isdst);
+#undef SET
+	if (PyErr_Occurred()) {
+		Py_XDECREF(v);
+		return NULL;
+	}
+
+	return v;
 }
 
 static PyObject *
@@ -256,7 +289,7 @@ time_gmtime(PyObject *self, PyObject *args)
 }
 
 static char gmtime_doc[] =
-"gmtime([seconds]) -> tuple\n\
+"gmtime([seconds]) -> (year,month,day,hour,minute,second,weekday,dayofyear,dst)\n\
 \n\
 Convert seconds since the Epoch to a time tuple expressing UTC (a.k.a.\n\
 GMT).  When 'seconds' is not passed in, convert the current time instead.";
@@ -273,7 +306,8 @@ time_localtime(PyObject *self, PyObject *args)
 }
 
 static char localtime_doc[] =
-"localtime([seconds]) -> tuple\n\
+"localtime([seconds]) -> (year,month,day,hour,minute,second,weekday,dayofyear,dst)\n\
+\n\
 Convert seconds since the Epoch to a time tuple expressing local time.\n\
 When 'seconds' is not passed in, convert the current time instead.";
 
@@ -309,7 +343,7 @@ gettmarg(PyObject *args, struct tm *p)
 			y += 2000;
 		else {
 			PyErr_SetString(PyExc_ValueError,
-					"year out of range (00-99, 1900-*)");
+					"year out of range");
 			return 0;
 		}
 	}
@@ -341,7 +375,7 @@ time_strftime(PyObject *self, PyObject *args)
 		buf = *localtime(&tt);
 	} else if (!gettmarg(tup, &buf))
 		return NULL;
-	
+
 	fmtlen = strlen(fmt);
 
 	/* I hate these functions that presume you know how big the output
@@ -411,6 +445,7 @@ time_strptime(PyObject *self, PyObject *args)
 
 static char strptime_doc[] =
 "strptime(string, format) -> tuple\n\
+\n\
 Parse a string to a time tuple according to a format specification.\n\
 See the library reference manual for formatting codes (same as strftime()).";
 #endif /* HAVE_STRPTIME */
@@ -447,7 +482,7 @@ time_ctime(PyObject *self, PyObject *args)
 	double dt;
 	time_t tt;
 	char *p;
-	
+
 	if (PyTuple_Size(args) == 0)
 		tt = time(NULL);
 	else {
@@ -491,7 +526,7 @@ time_mktime(PyObject *self, PyObject *args)
 	tt = mktime(&buf);
 	if (tt == (time_t)(-1)) {
 		PyErr_SetString(PyExc_OverflowError,
-                                "mktime argument out of range");
+				"mktime argument out of range");
 		return NULL;
 	}
 #if defined(macintosh) && defined(USE_GUSI211)
@@ -584,7 +619,7 @@ mktime() -- convert local time tuple to seconds since Epoch\n\
 strftime() -- convert time tuple to string according to format specification\n\
 strptime() -- parse string to time tuple according to format specification\n\
 ";
-  
+
 
 DL_EXPORT(void)
 inittime(void)
@@ -594,12 +629,12 @@ inittime(void)
 	m = Py_InitModule3("time", time_methods, module_doc);
 	d = PyModule_GetDict(m);
 	/* Accept 2-digit dates unless PYTHONY2K is set and non-empty */
-	p = getenv("PYTHONY2K");
+	p = Py_GETENV("PYTHONY2K");
 	ins(d, "accept2dyear", PyInt_FromLong((long) (!p || !*p)));
 	/* Squirrel away the module's dictionary for the y2k check */
 	Py_INCREF(d);
 	moddict = d;
-#if defined(HAVE_TZNAME) && !defined(__GLIBC__)
+#if defined(HAVE_TZNAME) && !defined(__GLIBC__) && !defined(__CYGWIN__)
 	tzset();
 #ifdef PYOS_OS2
 	ins(d, "timezone", PyInt_FromLong((long)_timezone));
@@ -617,7 +652,7 @@ inittime(void)
 #endif
 	ins(d, "daylight", PyInt_FromLong((long)daylight));
 	ins(d, "tzname", Py_BuildValue("(zz)", tzname[0], tzname[1]));
-#else /* !HAVE_TZNAME || __GLIBC__ */
+#else /* !HAVE_TZNAME || __GLIBC__ || __CYGWIN__*/
 #ifdef HAVE_TM_ZONE
 	{
 #define YEAR ((time_t)((365 * 24 + 6) * 3600))
@@ -635,7 +670,7 @@ inittime(void)
 		julyzone = -p->tm_gmtoff;
 		strncpy(julyname, p->tm_zone ? p->tm_zone : "   ", 9);
 		julyname[9] = '\0';
-		
+
 		if( janzone < julyzone ) {
 			/* DST is reversed in the southern hemisphere */
 			ins(d, "timezone", PyInt_FromLong(julyzone));
@@ -673,7 +708,10 @@ inittime(void)
 	ins(d, "daylight", PyInt_FromLong(_daylight));
 	ins(d, "tzname", Py_BuildValue("(zz)", _tzname[0], _tzname[1]));
 #endif /* __CYGWIN__ */
-#endif /* !HAVE_TZNAME || __GLIBC__ */
+#endif /* !HAVE_TZNAME || __GLIBC__ || __CYGWIN__*/
+
+        PyStructSequence_InitType(&StructTimeType, &struct_time_type_desc);
+	PyDict_SetItemString(d, "struct_time", (PyObject*) &StructTimeType);
 }
 
 
@@ -820,7 +858,7 @@ floatsleep(double secs)
 		if( secs <= 0.0 ) {
 			return;
 		}
-		
+
 		Py_BEGIN_ALLOW_THREADS
 		/* BeOS snooze() is in microseconds... */
 		if( snooze( (bigtime_t)( secs * 1000.0 * 1000.0 ) ) == B_INTERRUPTED ) {
@@ -852,5 +890,6 @@ floatsleep(double secs)
 #endif /* !__WATCOMC__ || __QNX__ */
 #endif /* !macintosh */
 #endif /* !HAVE_SELECT */
+
 	return 0;
 }
