@@ -4,6 +4,7 @@
 #include "Python.h"
 #include <ctype.h>
 #include "structmember.h" /* we need the offsetof() macro from there */
+#include "longintrepr.h"
 
 #define NEW_STYLE_NUMBER(o) PyType_HasFeature((o)->ob_type, \
 				Py_TPFLAGS_CHECKTYPES)
@@ -102,7 +103,8 @@ PyObject_GetItem(PyObject *o, PyObject *key)
 				return NULL;
 			return PySequence_GetItem(o, key_value);
 		}
-		return type_error("sequence index must be integer");
+		else if (o->ob_type->tp_as_sequence->sq_item)
+			return type_error("sequence index must be integer");
 	}
 
 	return type_error("unsubscriptable object");
@@ -130,8 +132,10 @@ PyObject_SetItem(PyObject *o, PyObject *key, PyObject *value)
 				return -1;
 			return PySequence_SetItem(o, key_value, value);
 		}
-		type_error("sequence index must be integer");
-		return -1;
+		else if (o->ob_type->tp_as_sequence->sq_ass_item) {
+			type_error("sequence index must be integer");
+			return -1;
+		}
 	}
 
 	type_error("object does not support item assignment");
@@ -160,8 +164,10 @@ PyObject_DelItem(PyObject *o, PyObject *key)
 				return -1;
 			return PySequence_DelItem(o, key_value);
 		}
-		type_error("sequence index must be integer");
-		return -1;
+		else if (o->ob_type->tp_as_sequence->sq_ass_item) {
+			type_error("sequence index must be integer");
+			return -1;
+		}
 	}
 
 	type_error("object does not support item deletion");
@@ -181,27 +187,37 @@ int PyObject_AsCharBuffer(PyObject *obj,
 		return -1;
 	}
 	pb = obj->ob_type->tp_as_buffer;
-	if ( pb == NULL ||
+	if (pb == NULL ||
 	     pb->bf_getcharbuffer == NULL ||
-	     pb->bf_getsegcount == NULL ) {
+	     pb->bf_getsegcount == NULL) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a character buffer object");
-		goto onError;
+		return -1;
 	}
-	if ( (*pb->bf_getsegcount)(obj,NULL) != 1 ) {
+	if ((*pb->bf_getsegcount)(obj,NULL) != 1) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a single-segment buffer object");
-		goto onError;
+		return -1;
 	}
-	len = (*pb->bf_getcharbuffer)(obj,0,&pp);
+	len = (*pb->bf_getcharbuffer)(obj, 0, &pp);
 	if (len < 0)
-		goto onError;
+		return -1;
 	*buffer = pp;
 	*buffer_len = len;
 	return 0;
+}
 
- onError:
-	return -1;
+int
+PyObject_CheckReadBuffer(PyObject *obj)
+{
+	PyBufferProcs *pb = obj->ob_type->tp_as_buffer;
+
+	if (pb == NULL ||
+	    pb->bf_getreadbuffer == NULL ||
+	    pb->bf_getsegcount == NULL ||
+	    (*pb->bf_getsegcount)(obj, NULL) != 1)
+		return 0;
+	return 1;
 }
 
 int PyObject_AsReadBuffer(PyObject *obj,
@@ -217,27 +233,24 @@ int PyObject_AsReadBuffer(PyObject *obj,
 		return -1;
 	}
 	pb = obj->ob_type->tp_as_buffer;
-	if ( pb == NULL ||
+	if (pb == NULL ||
 	     pb->bf_getreadbuffer == NULL ||
-	     pb->bf_getsegcount == NULL ) {
+	     pb->bf_getsegcount == NULL) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a readable buffer object");
-		goto onError;
+		return -1;
 	}
-	if ( (*pb->bf_getsegcount)(obj,NULL) != 1 ) {
+	if ((*pb->bf_getsegcount)(obj, NULL) != 1) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a single-segment buffer object");
-		goto onError;
+		return -1;
 	}
-	len = (*pb->bf_getreadbuffer)(obj,0,&pp);
+	len = (*pb->bf_getreadbuffer)(obj, 0, &pp);
 	if (len < 0)
-		goto onError;
+		return -1;
 	*buffer = pp;
 	*buffer_len = len;
 	return 0;
-
- onError:
-	return -1;
 }
 
 int PyObject_AsWriteBuffer(PyObject *obj,
@@ -253,27 +266,24 @@ int PyObject_AsWriteBuffer(PyObject *obj,
 		return -1;
 	}
 	pb = obj->ob_type->tp_as_buffer;
-	if ( pb == NULL ||
+	if (pb == NULL ||
 	     pb->bf_getwritebuffer == NULL ||
-	     pb->bf_getsegcount == NULL ) {
+	     pb->bf_getsegcount == NULL) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a writeable buffer object");
-		goto onError;
+		return -1;
 	}
-	if ( (*pb->bf_getsegcount)(obj,NULL) != 1 ) {
+	if ((*pb->bf_getsegcount)(obj, NULL) != 1) {
 		PyErr_SetString(PyExc_TypeError,
 				"expected a single-segment buffer object");
-		goto onError;
+		return -1;
 	}
 	len = (*pb->bf_getwritebuffer)(obj,0,&pp);
 	if (len < 0)
-		goto onError;
+		return -1;
 	*buffer = pp;
 	*buffer_len = len;
 	return 0;
-
- onError:
-	return -1;
 }
 
 /* Operations on numbers */
@@ -299,11 +309,14 @@ PyNumber_Check(PyObject *o)
 
   v	w	Action
   -------------------------------------------------------------------
-  new	new	v.op(v,w), w.op(v,w)
+  new	new	w.op(v,w)[*], v.op(v,w), w.op(v,w)
   new	old	v.op(v,w), coerce(v,w), v.op(v,w)
   old	new	w.op(v,w), coerce(v,w), v.op(v,w)
   old	old	coerce(v,w), v.op(v,w)
 
+  [*] only when v->ob_type != w->ob_type && w->ob_type is a subclass of
+      v->ob_type
+
   Legend:
   -------
   * new == new style number
@@ -317,29 +330,35 @@ static PyObject *
 binary_op1(PyObject *v, PyObject *w, const int op_slot)
 {
 	PyObject *x;
-	binaryfunc *slot;
-	if (v->ob_type->tp_as_number != NULL && NEW_STYLE_NUMBER(v)) {
-		slot = NB_BINOP(v->ob_type->tp_as_number, op_slot);
-		if (*slot) {
-			x = (*slot)(v, w);
-			if (x != Py_NotImplemented) {
-				return x;
-			}
-			Py_DECREF(x); /* can't do it */
-		}
-		if (v->ob_type == w->ob_type) {
-			goto binop_error;
-		}
+	binaryfunc slotv = NULL;
+	binaryfunc slotw = NULL;
+
+	if (v->ob_type->tp_as_number != NULL && NEW_STYLE_NUMBER(v))
+		slotv = *NB_BINOP(v->ob_type->tp_as_number, op_slot);
+	if (w->ob_type != v->ob_type &&
+	    w->ob_type->tp_as_number != NULL && NEW_STYLE_NUMBER(w)) {
+		slotw = *NB_BINOP(w->ob_type->tp_as_number, op_slot);
+		if (slotw == slotv)
+			slotw = NULL;
 	}
-	if (w->ob_type->tp_as_number != NULL && NEW_STYLE_NUMBER(w)) {
-		slot = NB_BINOP(w->ob_type->tp_as_number, op_slot);
-		if (*slot) {
-			x = (*slot)(v, w);
-			if (x != Py_NotImplemented) {
+	if (slotv) {
+		if (slotw && PyType_IsSubtype(w->ob_type, v->ob_type)) {
+			x = slotw(v, w);
+			if (x != Py_NotImplemented)
 				return x;
-			}
 			Py_DECREF(x); /* can't do it */
+			slotw = NULL;
 		}
+		x = slotv(v, w);
+		if (x != Py_NotImplemented)
+			return x;
+		Py_DECREF(x); /* can't do it */
+	}
+	if (slotw) {
+		x = slotw(v, w);
+		if (x != Py_NotImplemented)
+			return x;
+		Py_DECREF(x); /* can't do it */
 	}
 	if (!NEW_STYLE_NUMBER(v) || !NEW_STYLE_NUMBER(w)) {
 		int err = PyNumber_CoerceEx(&v, &w);
@@ -349,9 +368,10 @@ binary_op1(PyObject *v, PyObject *w, const int op_slot)
 		if (err == 0) {
 			PyNumberMethods *mv = v->ob_type->tp_as_number;
 			if (mv) {
-				slot = NB_BINOP(mv, op_slot);
-				if (*slot) {
-					PyObject *x = (*slot)(v, w);
+				binaryfunc slot;
+				slot = *NB_BINOP(mv, op_slot);
+				if (slot) {
+					PyObject *x = slot(v, w);
 					Py_DECREF(v);
 					Py_DECREF(w);
 					return x;
@@ -362,7 +382,6 @@ binary_op1(PyObject *v, PyObject *w, const int op_slot)
 			Py_DECREF(w);
 		}
 	}
-binop_error:
 	Py_INCREF(Py_NotImplemented);
 	return Py_NotImplemented;
 }
@@ -373,8 +392,12 @@ binary_op(PyObject *v, PyObject *w, const int op_slot, const char *op_name)
 	PyObject *result = binary_op1(v, w, op_slot);
 	if (result == Py_NotImplemented) {
 		Py_DECREF(Py_NotImplemented);
-		PyErr_Format(PyExc_TypeError, 
-				"unsupported operand type(s) for %s", op_name);
+		PyErr_Format(
+			PyExc_TypeError, 
+			"unsupported operand type(s) for %s: '%s' and '%s'",
+			op_name,
+			v->ob_type->tp_name,
+			w->ob_type->tp_name);
 		return NULL;
 	}
 	return result;
@@ -384,6 +407,8 @@ binary_op(PyObject *v, PyObject *w, const int op_slot, const char *op_name)
 /*
   Calling scheme used for ternary operations:
 
+  *** In some cases, w.op is called before v.op; see binary_op1. ***
+
   v	w	z	Action
   -------------------------------------------------------------------
   new	new	new	v.op(v,w,z), w.op(v,w,z), z.op(v,w,z)
@@ -415,51 +440,50 @@ ternary_op(PyObject *v,
 	   const char *op_name)
 {
 	PyNumberMethods *mv, *mw, *mz;
-	register PyObject *x = NULL;
-	register ternaryfunc *slot;
+	PyObject *x = NULL;
+	ternaryfunc slotv = NULL;
+	ternaryfunc slotw = NULL;
+	ternaryfunc slotz = NULL;
 	
 	mv = v->ob_type->tp_as_number;
-	if (mv != NULL && NEW_STYLE_NUMBER(v)) {
-		/* try v.op(v,w,z) */
-		slot = NB_TERNOP(mv, op_slot);
-		if (*slot) {
-			x = (*slot)(v, w, z);
-			if (x != Py_NotImplemented)
-				return x;
-			/* Can't do it... fall through */
-			Py_DECREF(x);
-		}
-		if (v->ob_type == w->ob_type &&
-				(z == Py_None || z->ob_type == v->ob_type)) {
-			goto ternary_error;
-		}
-	}
 	mw = w->ob_type->tp_as_number;
-	if (mw != NULL && NEW_STYLE_NUMBER(w)) {
-		/* try w.op(v,w,z) */
-		slot = NB_TERNOP(mw,op_slot);
-		if (*slot) {
-			x = (*slot)(v, w, z);
+	if (mv != NULL && NEW_STYLE_NUMBER(v))
+		slotv = *NB_TERNOP(mv, op_slot);
+	if (w->ob_type != v->ob_type &&
+	    mv != NULL && NEW_STYLE_NUMBER(w)) {
+		slotw = *NB_TERNOP(mw, op_slot);
+		if (slotw == slotv)
+			slotw = NULL;
+	}
+	if (slotv) {
+		if (slotw && PyType_IsSubtype(w->ob_type, v->ob_type)) {
+			x = slotw(v, w, z);
 			if (x != Py_NotImplemented)
 				return x;
-			/* Can't do it... fall through */
-			Py_DECREF(x);
-		}
-		if (NEW_STYLE_NUMBER(v) &&
-				(z == Py_None || z->ob_type == v->ob_type)) {
-			goto ternary_error;
+			Py_DECREF(x); /* can't do it */
+			slotw = NULL;
 		}
+		x = slotv(v, w, z);
+		if (x != Py_NotImplemented)
+			return x;
+		Py_DECREF(x); /* can't do it */
+	}
+	if (slotw) {
+		x = slotw(v, w, z);
+		if (x != Py_NotImplemented)
+			return x;
+		Py_DECREF(x); /* can't do it */
 	}
 	mz = z->ob_type->tp_as_number;
 	if (mz != NULL && NEW_STYLE_NUMBER(z)) {
-		/* try: z.op(v,w,z) */
-		slot = NB_TERNOP(mz, op_slot);
-		if (*slot) {
-			x = (*slot)(v, w, z);
+		slotz = *NB_TERNOP(mz, op_slot);
+		if (slotz == slotv || slotz == slotw)
+			slotz = NULL;
+		if (slotz) {
+			x = slotz(v, w, z);
 			if (x != Py_NotImplemented)
 				return x;
-			/* Can't do it... fall through */
-			Py_DECREF(x);
+			Py_DECREF(x); /* can't do it */
 		}
 	}
 
@@ -477,10 +501,10 @@ ternary_op(PyObject *v,
 		   treated as absent argument and not coerced. */
 		if (z == Py_None) {
 			if (v->ob_type->tp_as_number) {
-				slot = NB_TERNOP(v->ob_type->tp_as_number,
-						 op_slot);
-				if (*slot)
-					x = (*slot)(v, w, z);
+				slotz = *NB_TERNOP(v->ob_type->tp_as_number,
+						   op_slot);
+				if (slotz)
+					x = slotz(v, w, z);
 				else
 					c = -1;
 			}
@@ -500,10 +524,10 @@ ternary_op(PyObject *v,
 			goto error1;
 
 		if (v1->ob_type->tp_as_number != NULL) {
-			slot = NB_TERNOP(v1->ob_type->tp_as_number,
-					 op_slot);
-			if (*slot)
-				x = (*slot)(v1, w2, z2);
+			slotv = *NB_TERNOP(v1->ob_type->tp_as_number,
+					   op_slot);
+			if (slotv)
+				x = slotv(v1, w2, z2);
 			else
 				c = -1;
 		}
@@ -522,10 +546,22 @@ ternary_op(PyObject *v,
 		if (c >= 0)
 			return x;
 	}
-	
-ternary_error:
-	PyErr_Format(PyExc_TypeError, "unsupported operand type(s) for %s",
-			op_name);
+
+	if (z == Py_None)
+		PyErr_Format(
+			PyExc_TypeError,
+			"unsupported operand type(s) for ** or pow(): "
+			"'%s' and '%s'",
+			v->ob_type->tp_name,
+			w->ob_type->tp_name);
+	else
+		PyErr_Format(
+			PyExc_TypeError,
+			"unsupported operand type(s) for pow(): "
+			"'%s', '%s', '%s'",
+			v->ob_type->tp_name,
+			w->ob_type->tp_name,
+			z->ob_type->tp_name);
 	return NULL;
 }
 
@@ -556,21 +592,40 @@ PyNumber_Add(PyObject *v, PyObject *w)
 			result = (*m->sq_concat)(v, w);
 		}
                 else {
-                    PyErr_SetString(PyExc_TypeError,
-                                    "unsupported operand types for +");
+                    PyErr_Format(
+			    PyExc_TypeError,
+			    "unsupported operand types for +: '%s' and '%s'",
+			    v->ob_type->tp_name,
+			    w->ob_type->tp_name);
                     result = NULL;
                 }
 	}
 	return result;
 }
 
+PyObject *
+PyNumber_FloorDivide(PyObject *v, PyObject *w)
+{
+	/* XXX tp_flags test */
+	return binary_op(v, w, NB_SLOT(nb_floor_divide), "//");
+}
+
+PyObject *
+PyNumber_TrueDivide(PyObject *v, PyObject *w)
+{
+	/* XXX tp_flags test */
+	return binary_op(v, w, NB_SLOT(nb_true_divide), "/");
+}
+
 PyObject *
 PyNumber_Remainder(PyObject *v, PyObject *w)
 {
 	if (PyString_Check(v))
 		return PyString_Format(v, w);
+#ifdef Py_USING_UNICODE
 	else if (PyUnicode_Check(v))
 		return PyUnicode_Format(v, w);
+#endif
 	return binary_op(v, w, NB_SLOT(nb_remainder), "%");
 }
 
@@ -630,6 +685,22 @@ INPLACE_BINOP(PyNumber_InPlaceRshift, nb_inplace_rshift, nb_rshift, ">>=")
 INPLACE_BINOP(PyNumber_InPlaceSubtract, nb_inplace_subtract, nb_subtract, "-=")
 INPLACE_BINOP(PyNumber_InPlaceDivide, nb_inplace_divide, nb_divide, "/=")
 
+PyObject *
+PyNumber_InPlaceFloorDivide(PyObject *v, PyObject *w)
+{
+	/* XXX tp_flags test */
+	return binary_iop(v, w, NB_SLOT(nb_inplace_floor_divide),
+			  NB_SLOT(nb_floor_divide), "//=");
+}
+
+PyObject *
+PyNumber_InPlaceTrueDivide(PyObject *v, PyObject *w)
+{
+	/* XXX tp_flags test */
+	return binary_iop(v, w, NB_SLOT(nb_inplace_true_divide),
+			  NB_SLOT(nb_true_divide), "/=");
+}
+
 PyObject *
 PyNumber_InPlaceAdd(PyObject *v, PyObject *w)
 {
@@ -677,8 +748,10 @@ PyNumber_InPlaceRemainder(PyObject *v, PyObject *w)
 {
 	if (PyString_Check(v))
 		return PyString_Format(v, w);
+#ifdef Py_USING_UNICODE
 	else if (PyUnicode_Check(v))
 		return PyUnicode_Format(v, w);
+#endif
 	else
 		return binary_iop(v, w, NB_SLOT(nb_inplace_remainder),
 					NB_SLOT(nb_remainder), "%=");
@@ -784,17 +857,23 @@ PyNumber_Int(PyObject *o)
 
 	if (o == NULL)
 		return null_error();
-	if (PyInt_Check(o)) {
+	if (PyInt_CheckExact(o)) {
 		Py_INCREF(o);
 		return o;
 	}
+	if (PyInt_Check(o)) {
+		PyIntObject *io = (PyIntObject*)o;
+		return PyInt_FromLong(io->ob_ival);
+	}
 	if (PyString_Check(o))
 		return int_from_string(PyString_AS_STRING(o), 
 				       PyString_GET_SIZE(o));
+#ifdef Py_USING_UNICODE
 	if (PyUnicode_Check(o))
 		return PyInt_FromUnicode(PyUnicode_AS_UNICODE(o),
 					 PyUnicode_GET_SIZE(o),
 					 10);
+#endif
 	m = o->ob_type->tp_as_number;
 	if (m && m->nb_int)
 		return m->nb_int(o);
@@ -832,10 +911,12 @@ PyNumber_Long(PyObject *o)
 
 	if (o == NULL)
 		return null_error();
-	if (PyLong_Check(o)) {
+	if (PyLong_CheckExact(o)) {
 		Py_INCREF(o);
 		return o;
 	}
+	if (PyLong_Check(o))
+		return _PyLong_Copy((PyLongObject *)o);
 	if (PyString_Check(o))
 		/* need to do extra error checking that PyLong_FromString() 
 		 * doesn't do.  In particular long('9.5') must raise an
@@ -843,11 +924,13 @@ PyNumber_Long(PyObject *o)
 		 */
 		return long_from_string(PyString_AS_STRING(o),
 					PyString_GET_SIZE(o));
+#ifdef Py_USING_UNICODE
 	if (PyUnicode_Check(o))
 		/* The above check is done in PyLong_FromUnicode(). */
 		return PyLong_FromUnicode(PyUnicode_AS_UNICODE(o),
 					  PyUnicode_GET_SIZE(o),
 					  10);
+#endif
 	m = o->ob_type->tp_as_number;
 	if (m && m->nb_long)
 		return m->nb_long(o);
@@ -864,10 +947,14 @@ PyNumber_Float(PyObject *o)
 
 	if (o == NULL)
 		return null_error();
-	if (PyFloat_Check(o)) {
+	if (PyFloat_CheckExact(o)) {
 		Py_INCREF(o);
 		return o;
 	}
+	if (PyFloat_Check(o)) {
+		PyFloatObject *po = (PyFloatObject *)o;
+		return PyFloat_FromDouble(po->ob_fval);
+	}
 	if (!PyString_Check(o)) {
 		m = o->ob_type->tp_as_number;
 		if (m && m->nb_float)
@@ -881,7 +968,8 @@ PyNumber_Float(PyObject *o)
 int
 PySequence_Check(PyObject *s)
 {
-	return s != NULL && s->ob_type->tp_as_sequence;
+	return s != NULL && s->ob_type->tp_as_sequence &&
+		s->ob_type->tp_as_sequence->sq_item != NULL;
 }
 
 int
@@ -1176,112 +1264,147 @@ PySequence_DelSlice(PyObject *s, int i1, int i2)
 PyObject *
 PySequence_Tuple(PyObject *v)
 {
-	PySequenceMethods *m;
+	PyObject *it;  /* iter(v) */
+	int n;         /* guess for result tuple size */
+	PyObject *result;
+	int j;
 
 	if (v == NULL)
 		return null_error();
 
-	if (PyTuple_Check(v)) {
+	/* Special-case the common tuple and list cases, for efficiency. */
+	if (PyTuple_CheckExact(v)) {
+		/* Note that we can't know whether it's safe to return
+		   a tuple *subclass* instance as-is, hence the restriction
+		   to exact tuples here.  In contrast, lists always make
+		   a copy, so there's no need for exactness below. */
 		Py_INCREF(v);
 		return v;
 	}
-
 	if (PyList_Check(v))
 		return PyList_AsTuple(v);
 
-	/* There used to be code for strings here, but tuplifying strings is
-	   not a common activity, so I nuked it.  Down with code bloat! */
+	/* Get iterator. */
+	it = PyObject_GetIter(v);
+	if (it == NULL)
+		return NULL;
 
-	/* Generic sequence object */
-	m = v->ob_type->tp_as_sequence;
-	if (m && m->sq_item) {
-		int i;
-		PyObject *t;
-		int n = PySequence_Size(v);
-		if (n < 0)
-			return NULL;
-		t = PyTuple_New(n);
-		if (t == NULL)
-			return NULL;
-		for (i = 0; ; i++) {
-			PyObject *item = (*m->sq_item)(v, i);
-			if (item == NULL) {
-				if (PyErr_ExceptionMatches(PyExc_IndexError))
-					PyErr_Clear();
-				else {
-					Py_DECREF(t);
-					t = NULL;
-				}
-				break;
-			}
-			if (i >= n) {
-				if (n < 500)
-					n += 10;
-				else
-					n += 100;
-				if (_PyTuple_Resize(&t, n, 0) != 0)
-					break;
-			}
-			PyTuple_SET_ITEM(t, i, item);
+	/* Guess result size and allocate space. */
+	n = PySequence_Size(v);
+	if (n < 0) {
+		PyErr_Clear();
+		n = 10;  /* arbitrary */
+	}
+	result = PyTuple_New(n);
+	if (result == NULL)
+		goto Fail;
+
+	/* Fill the tuple. */
+	for (j = 0; ; ++j) {
+		PyObject *item = PyIter_Next(it);
+		if (item == NULL) {
+			if (PyErr_Occurred())
+				goto Fail;
+			break;
 		}
-		if (i < n && t != NULL)
-			_PyTuple_Resize(&t, i, 0);
-		return t;
+		if (j >= n) {
+			if (n < 500)
+				n += 10;
+			else
+				n += 100;
+			if (_PyTuple_Resize(&result, n) != 0) {
+				Py_DECREF(item);
+				goto Fail;
+			}
+		}
+		PyTuple_SET_ITEM(result, j, item);
 	}
 
-	/* None of the above */
-	return type_error("tuple() argument must be a sequence");
+	/* Cut tuple back if guess was too large. */
+	if (j < n &&
+	    _PyTuple_Resize(&result, j) != 0)
+		goto Fail;
+
+	Py_DECREF(it);
+	return result;
+
+Fail:
+	Py_XDECREF(result);
+	Py_DECREF(it);
+	return NULL;
 }
 
 PyObject *
 PySequence_List(PyObject *v)
 {
-	PySequenceMethods *m;
+	PyObject *it;      /* iter(v) */
+	PyObject *result;  /* result list */
+	int n;		   /* guess for result list size */
+	int i;
 
 	if (v == NULL)
 		return null_error();
 
+	/* Special-case list(a_list), for speed. */
 	if (PyList_Check(v))
 		return PyList_GetSlice(v, 0, PyList_GET_SIZE(v));
 
-	m = v->ob_type->tp_as_sequence;
-	if (m && m->sq_item) {
-		int i;
-		PyObject *l;
-		int n = PySequence_Size(v);
+	/* Get iterator.  There may be some low-level efficiency to be gained
+	 * by caching the tp_iternext slot instead of using PyIter_Next()
+	 * later, but premature optimization is the root etc.
+	 */
+	it = PyObject_GetIter(v);
+	if (it == NULL)
+		return NULL;
+
+	/* Guess a result list size. */
+	n = -1;	 /* unknown */
+	if (PySequence_Check(v) &&
+	    v->ob_type->tp_as_sequence->sq_length) {
+		n = PySequence_Size(v);
 		if (n < 0)
-			return NULL;
-		l = PyList_New(n);
-		if (l == NULL)
-			return NULL;
-		for (i = 0; ; i++) {
-			PyObject *item = (*m->sq_item)(v, i);
-			if (item == NULL) {
-				if (PyErr_ExceptionMatches(PyExc_IndexError))
-					PyErr_Clear();
-				else {
-					Py_DECREF(l);
-					l = NULL;
-				}
-				break;
-			}
-			if (i < n)
-				PyList_SET_ITEM(l, i, item);
-			else if (PyList_Append(l, item) < 0) {
-				Py_DECREF(l);
-				l = NULL;
-				break;
-			}
-		}
-		if (i < n && l != NULL) {
-			if (PyList_SetSlice(l, i, n, (PyObject *)NULL) != 0) {
-				Py_DECREF(l);
-				l = NULL;
-			}
-		}
-		return l;
+			PyErr_Clear();
 	}
-	return type_error("list() argument must be a sequence");
+	if (n < 0)
+		n = 8;	/* arbitrary */
+	result = PyList_New(n);
+	if (result == NULL) {
+		Py_DECREF(it);
+		return NULL;
+	}
+
+	/* Run iterator to exhaustion. */
+	for (i = 0; ; i++) {
+		PyObject *item = PyIter_Next(it);
+		if (item == NULL) {
+			if (PyErr_Occurred()) {
+				Py_DECREF(result);
+				result = NULL;
+			}
+			break;
+		}
+		if (i < n)
+			PyList_SET_ITEM(result, i, item); /* steals ref */
+		else {
+			int status = PyList_Append(result, item);
+			Py_DECREF(item);  /* append creates a new ref */
+			if (status < 0) {
+				Py_DECREF(result);
+				result = NULL;
+				break;
+			}
+		}
+	}
+
+	/* Cut back result list if initial guess was too large. */
+	if (i < n && result != NULL) {
+		if (PyList_SetSlice(result, i, n, (PyObject *)NULL) != 0) {
+			Py_DECREF(result);
+			result = NULL;
+		}
+	}
+	Py_DECREF(it);
+	return result;
 }
 
 PyObject *
@@ -1302,76 +1425,113 @@ PySequence_Fast(PyObject *v, const char *m)
 	return v;
 }
 
+/* Iterate over seq.  Result depends on the operation:
+   PY_ITERSEARCH_COUNT:  -1 if error, else # of times obj appears in seq.
+   PY_ITERSEARCH_INDEX:  0-based index of first occurence of obj in seq;
+   	set ValueError and return -1 if none found; also return -1 on error.
+   Py_ITERSEARCH_CONTAINS:  return 1 if obj in seq, else 0; -1 on error.
+*/
 int
-PySequence_Count(PyObject *s, PyObject *o)
+_PySequence_IterSearch(PyObject *seq, PyObject *obj, int operation)
 {
-	int l, i, n, cmp, err;
-	PyObject *item;
+	int n;
+	int wrapped;  /* for PY_ITERSEARCH_INDEX, true iff n wrapped around */
+	PyObject *it;  /* iter(seq) */
 
-	if (s == NULL || o == NULL) {
+	if (seq == NULL || obj == NULL) {
 		null_error();
 		return -1;
 	}
-	
-	l = PySequence_Size(s);
-	if (l < 0)
-		return -1;
 
-	n = 0;
-	for (i = 0; i < l; i++) {
-		item = PySequence_GetItem(s, i);
-		if (item == NULL)
-			return -1;
-		err = PyObject_Cmp(item, o, &cmp);
-		Py_DECREF(item);
-		if (err < 0)
-			return err;
-		if (cmp == 0)
-			n++;
+	it = PyObject_GetIter(seq);
+	if (it == NULL) {
+		type_error("iterable argument required");
+		return -1;
 	}
+
+	n = wrapped = 0;
+	for (;;) {
+		int cmp;
+		PyObject *item = PyIter_Next(it);
+		if (item == NULL) {
+			if (PyErr_Occurred())
+				goto Fail;
+			break;
+		}
+
+		cmp = PyObject_RichCompareBool(obj, item, Py_EQ);
+		Py_DECREF(item);
+		if (cmp < 0)
+			goto Fail;
+		if (cmp > 0) {
+			switch (operation) {
+			case PY_ITERSEARCH_COUNT:
+				++n;
+				if (n <= 0) {
+					PyErr_SetString(PyExc_OverflowError,
+				                "count exceeds C int size");
+					goto Fail;
+				}
+				break;
+
+			case PY_ITERSEARCH_INDEX:
+				if (wrapped) {
+					PyErr_SetString(PyExc_OverflowError,
+			                	"index exceeds C int size");
+					goto Fail;
+				}
+				goto Done;
+
+			case PY_ITERSEARCH_CONTAINS:
+				n = 1;
+				goto Done;
+
+			default:
+				assert(!"unknown operation");
+			}
+		}
+
+		if (operation == PY_ITERSEARCH_INDEX) {
+			++n;
+			if (n <= 0)
+				wrapped = 1;
+		}
+	}
+
+	if (operation != PY_ITERSEARCH_INDEX)
+		goto Done;
+
+	PyErr_SetString(PyExc_ValueError,
+		        "sequence.index(x): x not in sequence");
+	/* fall into failure code */
+Fail:
+	n = -1;
+	/* fall through */
+Done:
+	Py_DECREF(it);
 	return n;
+
 }
 
+/* Return # of times o appears in s. */
 int
-PySequence_Contains(PyObject *w, PyObject *v) /* v in w */
+PySequence_Count(PyObject *s, PyObject *o)
 {
-	int i, cmp;
-	PyObject *x;
-	PySequenceMethods *sq;
+	return _PySequence_IterSearch(s, o, PY_ITERSEARCH_COUNT);
+}
 
-	if(PyType_HasFeature(w->ob_type, Py_TPFLAGS_HAVE_SEQUENCE_IN)) {
-		sq = w->ob_type->tp_as_sequence;
-	        if(sq != NULL && sq->sq_contains != NULL)
-			return (*sq->sq_contains)(w, v);
+/* Return -1 if error; 1 if ob in seq; 0 if ob not in seq.
+ * Use sq_contains if possible, else defer to _PySequence_IterSearch().
+ */
+int
+PySequence_Contains(PyObject *seq, PyObject *ob)
+{
+	if (PyType_HasFeature(seq->ob_type, Py_TPFLAGS_HAVE_SEQUENCE_IN)) {
+		PySequenceMethods *sqm = seq->ob_type->tp_as_sequence;
+	        if (sqm != NULL && sqm->sq_contains != NULL)
+			return (*sqm->sq_contains)(seq, ob);
 	}
-	
-	/* If there is no better way to check whether an item is is contained,
-	   do it the hard way */
-	sq = w->ob_type->tp_as_sequence;
-	if (sq == NULL || sq->sq_item == NULL) {
-		PyErr_SetString(PyExc_TypeError,
-			"'in' or 'not in' needs sequence right argument");
-		return -1;
-	}
-
-	for (i = 0; ; i++) {
-		x = (*sq->sq_item)(w, i);
-		if (x == NULL) {
-			if (PyErr_ExceptionMatches(PyExc_IndexError)) {
-				PyErr_Clear();
-				break;
-			}
-			return -1;
-		}
-		cmp = PyObject_RichCompareBool(v, x, Py_EQ);
-		Py_XDECREF(x);
-		if (cmp > 0)
-			return 1;
-		if (cmp < 0)
-			return -1;
-	}
-
-	return 0;
+	return _PySequence_IterSearch(seq, ob, PY_ITERSEARCH_CONTAINS);
 }
 
 /* Backwards compatibility */
@@ -1385,32 +1545,7 @@ PySequence_In(PyObject *w, PyObject *v)
 int
 PySequence_Index(PyObject *s, PyObject *o)
 {
-	int l, i, cmp, err;
-	PyObject *item;
-
-	if (s == NULL || o == NULL) {
-		null_error();
-		return -1;
-	}
-	
-	l = PySequence_Size(s);
-	if (l < 0)
-		return -1;
-
-	for (i = 0; i < l; i++) {
-		item = PySequence_GetItem(s, i);
-		if (item == NULL)
-			return -1;
-		err = PyObject_Cmp(item, o, &cmp);
-		Py_DECREF(item);
-		if (err < 0)
-			return err;
-		if (cmp == 0)
-			return i;
-	}
-
-	PyErr_SetString(PyExc_ValueError, "sequence.index(x): x not in list");
-	return -1;
+	return _PySequence_IterSearch(s, o, PY_ITERSEARCH_INDEX);
 }
 
 /* Operations on mappings */
@@ -1418,7 +1553,8 @@ PySequence_Index(PyObject *s, PyObject *o)
 int
 PyMapping_Check(PyObject *o)
 {
-	return o && o->ob_type->tp_as_mapping;
+	return o && o->ob_type->tp_as_mapping &&
+		o->ob_type->tp_as_mapping->mp_subscript;
 }
 
 int
@@ -1517,22 +1653,25 @@ PyMapping_HasKey(PyObject *o, PyObject *key)
 PyObject *
 PyObject_CallObject(PyObject *o, PyObject *a)
 {
-	PyObject *r;
-	PyObject *args = a;
+	return PyEval_CallObjectWithKeywords(o, a, NULL);
+}
 
-	if (args == NULL) {
-		args = PyTuple_New(0);
-		if (args == NULL)
-			return NULL;
+PyObject *
+PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw)
+{
+        ternaryfunc call;
+
+	if ((call = func->ob_type->tp_call) != NULL) {
+		PyObject *result = (*call)(func, arg, kw);
+		if (result == NULL && !PyErr_Occurred())
+			PyErr_SetString(
+				PyExc_SystemError,
+				"NULL result without error in PyObject_Call");
+		return result;
 	}
-
-	r = PyEval_CallObject(o, args);
-
-	if (args != a) {
-		Py_DECREF(args);
-	}
-
-	return r;
+	PyErr_Format(PyExc_TypeError, "'%s' object is not callable",
+		     func->ob_type->tp_name);
+	return NULL;
 }
 
 PyObject *
@@ -1540,20 +1679,18 @@ PyObject_CallFunction(PyObject *callable, char *format, ...)
 {
 	va_list va;
 	PyObject *args, *retval;
-	va_start(va, format);
 
-	if (callable == NULL) {
-		va_end(va);
+	if (callable == NULL)
 		return null_error();
-	}
 
-	if (format)
+	if (format && *format) {
+		va_start(va, format);
 		args = Py_VaBuildValue(format, va);
+		va_end(va);
+	}
 	else
 		args = PyTuple_New(0);
 
-	va_end(va);
-	
 	if (args == NULL)
 		return NULL;
 
@@ -1579,32 +1716,27 @@ PyObject_CallMethod(PyObject *o, char *name, char *format, ...)
 {
 	va_list va;
 	PyObject *args, *func = 0, *retval;
-	va_start(va, format);
 
-	if (o == NULL || name == NULL) {
-		va_end(va);
+	if (o == NULL || name == NULL)
 		return null_error();
-	}
 
 	func = PyObject_GetAttrString(o, name);
 	if (func == NULL) {
-		va_end(va);
 		PyErr_SetString(PyExc_AttributeError, name);
 		return 0;
 	}
 
-	if (!PyCallable_Check(func)) {
-		va_end(va);
+	if (!PyCallable_Check(func))
 		return type_error("call of non-callable attribute");
-	}
 
-	if (format && *format)
+	if (format && *format) {
+		va_start(va, format);
 		args = Py_VaBuildValue(format, va);
+		va_end(va);
+	}
 	else
 		args = PyTuple_New(0);
 
-	va_end(va);
-
 	if (!args)
 		return NULL;
 
@@ -1628,47 +1760,124 @@ PyObject_CallMethod(PyObject *o, char *name, char *format, ...)
 }
 
 
+static PyObject *
+objargs_mktuple(va_list va)
+{
+	int i, n = 0;
+	va_list countva;
+	PyObject *result, *tmp;
+
+#ifdef VA_LIST_IS_ARRAY
+	memcpy(countva, va, sizeof(va_list));
+#else
+	countva = va;
+#endif
+
+	while (((PyObject *)va_arg(countva, PyObject *)) != NULL)
+		++n;
+	result = PyTuple_New(n);
+	if (result != NULL && n > 0) {
+		for (i = 0; i < n; ++i) {
+			tmp = (PyObject *)va_arg(va, PyObject *);
+			PyTuple_SET_ITEM(result, i, tmp);
+			Py_INCREF(tmp);
+		}
+	}
+	return result;
+}
+
+PyObject *
+PyObject_CallMethodObjArgs(PyObject *callable, PyObject *name, ...)
+{
+	PyObject *args, *tmp;
+	va_list vargs;
+
+	if (callable == NULL || name == NULL)
+		return null_error();
+
+	callable = PyObject_GetAttr(callable, name);
+	if (callable == NULL)
+		return NULL;
+
+	/* count the args */
+	va_start(vargs, name);
+	args = objargs_mktuple(vargs);
+	va_end(vargs);
+	if (args == NULL) {
+		Py_DECREF(callable);
+		return NULL;
+	}
+	tmp = PyObject_Call(callable, args, NULL);
+	Py_DECREF(args);
+	Py_DECREF(callable);
+
+	return tmp;
+}
+
+PyObject *
+PyObject_CallFunctionObjArgs(PyObject *callable, ...)
+{
+	PyObject *args, *tmp;
+	va_list vargs;
+
+	if (callable == NULL)
+		return null_error();
+
+	/* count the args */
+	va_start(vargs, callable);
+	args = objargs_mktuple(vargs);
+	va_end(vargs);
+	if (args == NULL)
+		return NULL;
+	tmp = PyObject_Call(callable, args, NULL);
+	Py_DECREF(args);
+
+	return tmp;
+}
+
+
 /* isinstance(), issubclass() */
 
-static int
-abstract_issubclass(PyObject *derived, PyObject *cls, int first)
+static PyObject *
+abstract_get_bases(PyObject *cls)
 {
 	static PyObject *__bases__ = NULL;
 	PyObject *bases;
-	int i, n;
-	int r = 0;
 
 	if (__bases__ == NULL) {
 		__bases__ = PyString_FromString("__bases__");
 		if (__bases__ == NULL)
-			return -1;
+			return NULL;
 	}
 
-	if (first) {
-		bases = PyObject_GetAttr(cls, __bases__);
-		if (bases == NULL || !PyTuple_Check(bases)) {
-			Py_XDECREF(bases);
-			PyErr_SetString(PyExc_TypeError,
-					"issubclass() arg 2 must be a class");
-			return -1;
-		}
-		Py_DECREF(bases);
+	bases = PyObject_GetAttr(cls, __bases__);
+	if (bases == NULL || !PyTuple_Check(bases)) {
+	        Py_XDECREF(bases);
+		return NULL;
 	}
 
+	return bases;
+}
+
+
+static int
+abstract_issubclass(PyObject *derived, PyObject *cls)
+{
+	PyObject *bases;
+	int i, n;
+	int r = 0;
+
+
 	if (derived == cls)
 		return 1;
 
-	bases = PyObject_GetAttr(derived, __bases__);
-	if (bases == NULL || !PyTuple_Check(bases)) {
-	        Py_XDECREF(bases);
-		PyErr_SetString(PyExc_TypeError,
-				"issubclass() arg 1 must be a class");
-		return -1;
-	}
+	bases = abstract_get_bases(derived);
+	if (bases == NULL)
+		return 0;
 
 	n = PyTuple_GET_SIZE(bases);
 	for (i = 0; i < n; i++) {
-		r = abstract_issubclass(PyTuple_GET_ITEM(bases, i), cls, 0);
+		r = abstract_issubclass(PyTuple_GET_ITEM(bases, i), cls);
 		if (r != 0)
 			break;
 	}
@@ -1685,40 +1894,52 @@ PyObject_IsInstance(PyObject *inst, PyObject *cls)
 	static PyObject *__class__ = NULL;
 	int retval = 0;
 
-        if (PyClass_Check(cls)) {
-		if (PyInstance_Check(inst)) {
-			PyObject *inclass =
-				(PyObject*)((PyInstanceObject*)inst)->in_class;
-			retval = PyClass_IsSubclass(inclass, cls);
-		}
+	if (PyClass_Check(cls) && PyInstance_Check(inst)) {
+		PyObject *inclass =
+			(PyObject*)((PyInstanceObject*)inst)->in_class;
+		retval = PyClass_IsSubclass(inclass, cls);
 	}
 	else if (PyType_Check(cls)) {
-		retval = ((PyObject *)(inst->ob_type) == cls);
+		retval = PyObject_TypeCheck(inst, (PyTypeObject *)cls);
 	}
-	else if (!PyInstance_Check(inst)) {
+	else if (PyTuple_Check(cls)) {
+		/* Not a general sequence -- that opens up the road to
+		   recursion and stack overflow. */
+		int i, n;
+
+		n = PyTuple_GET_SIZE(cls);
+		for (i = 0; i < n; i++) {
+			retval = PyObject_IsInstance(
+				inst, PyTuple_GET_ITEM(cls, i));
+			if (retval != 0)
+				break;
+		}
+		return retval;
+	}
+	else {
+		PyObject *cls_bases = abstract_get_bases(cls);
+		if (cls_bases == NULL) {
+			PyErr_SetString(PyExc_TypeError, 
+				"isinstance() arg 2 must be a class or type");
+			return -1;
+		} 
+		Py_DECREF(cls_bases);
 		if (__class__ == NULL) {
 			__class__ = PyString_FromString("__class__");
 			if (__class__ == NULL)
 				return -1;
 		}
 		icls = PyObject_GetAttr(inst, __class__);
-		if (icls != NULL) {
-			retval = abstract_issubclass(icls, cls, 1);
+		if (icls == NULL) {
+			PyErr_Clear();
+			retval = 0;
+		}
+		else {
+			retval = abstract_issubclass(icls, cls);
 			Py_DECREF(icls);
-			if (retval < 0 &&
-			    !PyErr_ExceptionMatches(PyExc_TypeError))
-				return -1;
 		}
-		else
-			retval = -1;
 	}
-	else
-		retval = -1;
 
-	if (retval < 0) {
-		PyErr_SetString(PyExc_TypeError,
-				"isinstance() arg 2 must be a class or type");
-	}
 	return retval;
 }
 
@@ -1728,7 +1949,26 @@ PyObject_IsSubclass(PyObject *derived, PyObject *cls)
 	int retval;
 
 	if (!PyClass_Check(derived) || !PyClass_Check(cls)) {
-		retval = abstract_issubclass(derived, cls, 1);
+		PyObject *derived_bases;
+		PyObject *cls_bases;
+	       
+		derived_bases = abstract_get_bases(derived);
+		if (derived_bases == NULL) {
+			PyErr_SetString(PyExc_TypeError, 
+					"issubclass() arg 1 must be a class");
+			return -1;
+		}
+		Py_DECREF(derived_bases);
+
+		cls_bases = abstract_get_bases(cls);
+		if (cls_bases == NULL) {
+			PyErr_SetString(PyExc_TypeError, 
+					"issubclass() arg 2 must be a class");
+			return -1;
+		}
+		Py_DECREF(cls_bases);
+
+		retval = abstract_issubclass(derived, cls);
 	}
 	else {
 		/* shortcut */
@@ -1738,3 +1978,57 @@ PyObject_IsSubclass(PyObject *derived, PyObject *cls)
 
 	return retval;
 }
+
+PyObject *
+PyObject_GetIter(PyObject *o)
+{
+	PyTypeObject *t = o->ob_type;
+	getiterfunc f = NULL;
+	if (PyType_HasFeature(t, Py_TPFLAGS_HAVE_ITER))
+		f = t->tp_iter;
+	if (f == NULL) {
+		if (PySequence_Check(o))
+			return PySeqIter_New(o);
+		PyErr_SetString(PyExc_TypeError, 
+				"iteration over non-sequence");
+		return NULL;
+	}
+	else {
+		PyObject *res = (*f)(o);
+		if (res != NULL && !PyIter_Check(res)) {
+			PyErr_Format(PyExc_TypeError,
+				     "iter() returned non-iterator "
+				     "of type '%.100s'",
+				     res->ob_type->tp_name);
+			Py_DECREF(res);
+			res = NULL;
+		}
+		return res;
+	}
+}
+
+/* Return next item.
+ * If an error occurs, return NULL.  PyErr_Occurred() will be true.
+ * If the iteration terminates normally, return NULL and clear the
+ * PyExc_StopIteration exception (if it was set).  PyErr_Occurred()
+ * will be false.
+ * Else return the next object.  PyErr_Occurred() will be false.
+ */
+PyObject *
+PyIter_Next(PyObject *iter)
+{
+	PyObject *result;
+	if (!PyIter_Check(iter)) {
+		PyErr_Format(PyExc_TypeError,
+			     "'%.100s' object is not an iterator",
+			     iter->ob_type->tp_name);
+		return NULL;
+	}
+	result = (*iter->ob_type->tp_iternext)(iter);
+	if (result == NULL &&
+	    PyErr_Occurred() &&
+	    PyErr_ExceptionMatches(PyExc_StopIteration))
+		PyErr_Clear();
+	return result;
+}
+
